import torch
from curobo.cuda_robot_model.cuda_robot_model import CudaRobotModel
from curobo.geom.types import WorldConfig
from curobo.types.base import TensorDeviceType
from curobo.types.robot import RobotConfig
from curobo.util_file import (
    get_robot_path,
    get_world_configs_path,
    join_path,
    load_yaml,
)
from curobo.wrap.reacher.ik_solver import IKSolver, IKSolverConfig
from pytorch3d import transforms

from metasim.cfg.robots.base_robot_metacfg import BaseRobotMetaCfg


def get_curobo_models(robot_cfg: BaseRobotMetaCfg, no_gnd=False):
    tensor_args = TensorDeviceType()
    config_file = load_yaml(join_path(get_robot_path(), robot_cfg.curobo_ref_cfg_name))["robot_cfg"]
    curobo_robot_cfg = RobotConfig.from_dict(config_file, tensor_args)
    world_cfg = WorldConfig.from_dict(load_yaml(join_path(get_world_configs_path(), "collision_base.yml")))
    ik_config = IKSolverConfig.load_from_robot_config(
        curobo_robot_cfg,
        None if no_gnd else world_cfg,
        rotation_threshold=0.05,
        position_threshold=0.005,
        num_seeds=20,
        self_collision_check=True,
        self_collision_opt=True,
        tensor_args=tensor_args,
        use_cuda_graph=True,
    )

    ik_solver = IKSolver(ik_config)
    kin_model = CudaRobotModel(curobo_robot_cfg.kinematics)

    def do_fk(q: torch.Tensor):
        robot_state = kin_model.get_state(q, config_file["kinematics"]["ee_link"])
        return robot_state.ee_position, robot_state.ee_quaternion

    return kin_model, do_fk, ik_solver


# TODO: Add relative rot; currently all are 0
def ee_pose_from_tcp_pose(robot_cfg: BaseRobotMetaCfg, tcp_pos: torch.Tensor, tcp_quat: torch.Tensor):
    tcp_rel_pos = torch.tensor(robot_cfg.curobo_tcp_rel_pos).unsqueeze(0).to(tcp_pos.device)
    ee_pos = tcp_pos + torch.matmul(transforms.quaternion_to_matrix(tcp_quat), -tcp_rel_pos.unsqueeze(-1)).squeeze()
    return ee_pos, tcp_quat


def tcp_pose_from_ee_pose(robot_cfg: BaseRobotMetaCfg, ee_pos: torch.Tensor, ee_quat: torch.Tensor):
    ee_rotmat = transforms.quaternion_to_matrix(ee_quat)
    tcp_rel_pos = torch.tensor(robot_cfg.curobo_tcp_rel_pos).unsqueeze(0).to(ee_rotmat.device)
    tcp_pos = ee_pos + torch.matmul(ee_rotmat, tcp_rel_pos.unsqueeze(-1)).squeeze()
    return tcp_pos, ee_quat



from metasim.cfg.objects import ArticulationObjMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class PutShoesInBoxMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 600
    traj_filepath = "roboverse_data/trajs/rlbench/put_shoes_in_box/v2"
    objects = [
        ArticulationObjMetaCfg(
            name="box_base",
            filepath="roboverse_data/assets/rlbench/put_shoes_in_box/box_base/usd/box_base.usd",
        ),
        RigidObjMetaCfg(
            name="shoe1_visual",
            filepath="roboverse_data/assets/rlbench/put_shoes_in_box/shoe1_visual/usd/shoe1_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="shoe2_visual",
            filepath="roboverse_data/assets/rlbench/put_shoes_in_box/shoe2_visual/usd/shoe2_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



# ruff: noqa: F401
from .basketball_in_hoop_metacfg import BasketballInHoopMetaCfg
from .beat_the_buzz_metacfg import BeatTheBuzzMetaCfg
from .block_pyramid_metacfg import BlockPyramidMetaCfg
from .change_channel_metacfg import ChangeChannelMetaCfg
from .change_clock_metacfg import ChangeClockMetaCfg
from .close_box_metacfg import CloseBoxMetaCfg
from .close_door_metacfg import CloseDoorMetaCfg, OpenDoorMetaCfg
from .close_drawer_metacfg import CloseDrawerMetaCfg, OpenDrawerMetaCfg
from .close_fridge_metacfg import CloseFridgeMetaCfg, OpenFridgeMetaCfg
from .close_grill_metacfg import CloseGrillMetaCfg
from .close_jar_metacfg import CloseJarMetaCfg
from .close_laptop_lid_metacfg import CloseLaptopLidMetaCfg
from .close_microwave_metacfg import CloseMicrowaveMetaCfg, OpenMicrowaveMetaCfg
from .empty_dishwasher_metacfg import EmptyDishwasherMetaCfg
from .get_ice_from_fridge_metacfg import GetIceFromFridgeMetaCfg
from .hang_frame_on_hanger_metacfg import HangFrameOnHangerMetaCfg
from .hit_ball_with_queue_metacfg import HitBallWithQueueMetaCfg
from .hockey_metacfg import HockeyMetaCfg
from .insert_onto_square_peg_metacfg import InsertOntoSquarePegMetaCfg
from .lamp_off_metacfg import LampOffMetaCfg
from .lamp_on_metacfg import LampOnMetaCfg
from .lift_numbered_block_metacfg import LiftNumberedBlockMetaCfg
from .light_bulb_in_metacfg import LightBulbInMetaCfg
from .light_bulb_out_metacfg import LightBulbOutMetaCfg
from .meat_off_grill_metacfg import MeatOffGrillMetaCfg
from .meat_on_grill_metacfg import MeatOnGrillMetaCfg
from .open_box_metacfg import OpenBoxMetaCfg
from .open_grill_metacfg import OpenGrillMetaCfg
from .open_jar_metacfg import OpenJarMetaCfg
from .open_oven_metacfg import OpenOvenMetaCfg, PutTrayInOvenMetaCfg, TakeTrayOutOfOvenMetaCfg
from .open_washing_machine_metacfg import OpenWashingMachineMetaCfg
from .open_window_metacfg import OpenWindowMetaCfg
from .open_wine_bottle_metacfg import OpenWineBottleMetaCfg
from .phone_on_base_metacfg import PhoneOnBaseMetaCfg
from .pick_and_lift_metacfg import PickAndLiftMetaCfg
from .pick_and_lift_small_metacfg import PickAndLiftSmallMetaCfg
from .pick_up_cup_metacfg import PickUpCupMetaCfg
from .place_cups_metacfg import PlaceCupsMetaCfg
from .place_shape_in_shape_sorter_metacfg import PlaceShapeInShapeSorterMetaCfg
from .play_jenga_metacfg import PlayJengaMetaCfg
from .plug_charger_in_power_supply_metacfg import PlugChargerInPowerSupplyMetaCfg
from .pour_from_cup_to_cup_metacfg import PourFromCupToCupMetaCfg
from .press_switch_metacfg import PressSwitchMetaCfg
from .push_button_metacfg import PushButtonMetaCfg
from .push_buttons_metacfg import PushButtonsMetaCfg
from .put_all_groceries_in_cupboard_metacfg import PutAllGroceriesInCupboardMetaCfg
from .put_books_on_bookshelf_metacfg import PutBooksOnBookshelfMetaCfg
from .put_bottle_in_fridge_metacfg import PutBottleInFridgeMetaCfg
from .put_groceries_in_cupboard_metacfg import PutGroceriesInCupboardMetaCfg
from .put_item_in_drawer_metacfg import PutItemInDrawerMetaCfg
from .put_knife_in_knife_block_metacfg import PutKnifeInKnifeBlockMetaCfg, PutKnifeOnChoppingBoardMetaCfg
from .put_money_in_safe_metacfg import PutMoneyInSafeMetaCfg
from .put_plate_in_colored_dish_rack_metacfg import PutPlateInColoredDishRackMetaCfg
from .put_rubbish_in_bin_metacfg import PutRubbishInBinMetaCfg
from .put_shoes_in_box_metacfg import PutShoesInBoxMetaCfg
from .put_toilet_roll_on_stand_metacfg import PutToiletRollOnStandMetaCfg
# TODO: add task empty_container which uses lots of "procedural_objects"



# ruff: noqa: F401

from .base_scene_metacfg import SceneMetaCfg
from .tapwater_scene_131_metacfg import TapwaterScene131MetaCfg
from .tapwater_scene_138_metacfg import TapwaterScene138MetaCfg
from .tapwater_scene_144_metacfg import TapwaterScene144MetaCfg
from .tapwater_scene_152_metacfg import TapwaterScene152MetaCfg
from .tapwater_scene_155_metacfg import TapwaterScene155MetaCfg



from __future__ import annotations

import torch
from dm_control.utils import rewards

from metasim.cfg.checkers import StairChecker
from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.types import EnvState
from metasim.utils import configclass
from metasim.utils.h1_utils import (
    actuator_forces,
    center_of_mass_velocity,
    head_height,
    left_foot_height,
    right_foot_height,
    torso_upright,
)

from .base_locomotion_env import (
    _G1_CRAWL_HEIGHT,
    _G1_STAND_HEIGHT,
    _H1_CRAWL_HEIGHT,
    _H1_STAND_HEIGHT,
    HumanoidTaskMetaCfg,
)


class StairReward:
    def __init__(self, robot_name="h1"):
        self._robot_name = robot_name
        if robot_name == "h1" or robot_name == "h1_simple_hand":
            self._stand_height = _H1_STAND_HEIGHT
            self._crawl_height = _H1_CRAWL_HEIGHT
        elif robot_name == "g1":
            self._stand_height = _G1_STAND_HEIGHT
            self._crawl_height = _G1_CRAWL_HEIGHT
        else:
            raise ValueError(f"Unknown robot {robot_name}")

    def __call__(self, states: list[EnvState]) -> torch.FloatTensor:
        results = []
        for state in states:
            # Reward for controlled movement
            small_control = rewards.tolerance(
                actuator_forces(state, self._robot_name),
                margin=10,
                value_at_margin=0,
                sigmoid="quadratic",
            ).mean()
            small_control = (4 + small_control) / 5

            # Reward for sliding motion - encourage horizontal velocity
            root_x_speed = rewards.tolerance(
                center_of_mass_velocity(state, self._robot_name)[0],  # x-axis velocity
                bounds=(1.0, float("inf")),  # Adjust velocity threshold as needed
                margin=1.0,
                value_at_margin=0,
                sigmoid="linear",
            )

            # Reward for maintaining upright posture
            upright = rewards.tolerance(
                torso_upright(state, self._robot_name),
                bounds=(0.5, 1.0),
                sigmoid="linear",
                margin=1.9,
                value_at_margin=0,
            )

            # Left foot vertical distance
            head_z = head_height(state, self._robot_name)
            left_foot_z = left_foot_height(state, self._robot_name)
            right_foot_z = right_foot_height(state, self._robot_name)
            vertical_foot_left = rewards.tolerance(
                head_z - left_foot_z,
                bounds=(1.2, float("inf")),
                margin=0.45,
            )
            vertical_foot_right = rewards.tolerance(
                head_z - right_foot_z,
                bounds=(1.2, float("inf")),
                margin=0.45,
            )

            # Combine rewards
            reward = small_control * root_x_speed * upright * vertical_foot_left * vertical_foot_right

            # log.info(f"root_x_speed: {root_x_speed}, upright: {upright}, vertical_foot_left: {vertical_foot_left}, vertical_foot_right: {vertical_foot_right}")
            results.append(reward)
        return torch.tensor(results)


@configclass
class StairMetaCfg(HumanoidTaskMetaCfg):
    episode_length = 1000
    objects = [
        RigidObjMetaCfg(
            name="stair",
            mjcf_path="roboverse_data/assets/humanoidbench/stair/floor/mjcf/floor.xml",
            physics=PhysicStateType.GEOM,
            fix_base_link=True,
        ),
    ]
    traj_filepath = "roboverse_data/trajs/humanoidbench/stair/v2/initial_state_v2.json"
    checker = StairChecker()
    reward_weights = [1.0]
    reward_functions = [StairReward]



import torch
from loguru import logger as log

from metasim.cfg.objects import ArticulationObjMetaCfg, PrimitiveCubeMetaCfg, PrimitiveSphereMetaCfg, RigidObjMetaCfg
from metasim.cfg.scenario import ScenarioMetaCfg
from metasim.utils.camera_util import get_cam_params

from .isaaclab_helper import add_objects, add_robot, get_pose

try:
    from .empty_env import EmptyEnv
except:
    pass

## Constants

table_height = 0.75
table_thickness = 0.05
table_size = 1.5

wall_dist = 3.0
wall_height = 4.0
wall_thickness = 0.1

SCENE_PRIM_PATH = "/World/envs/env_0/house"


################################################################################
## Since IsaacLab is a OOP style framework, we need to overwrite some methods
################################################################################
class IsaaclabEnvOverwriter:
    def __init__(self, scenario: ScenarioMetaCfg):
        self.scenario = scenario
        self.task = scenario.task
        self.robot = scenario.robot
        self.cameras = scenario.cameras
        self.objects = scenario.task.objects
        self.scene = scenario.scene
        self.checker = scenario.task.checker
        self.checker_debug_viewers = self.checker.get_debug_viewers()

        ## XXX for initialization
        self.first_reset = True

    def _reset_idx(self, env: "EmptyEnv", env_ids: list[int] | None = None) -> None:
        if self.first_reset:
            log.info("Resetting for the first time, initializing randomizer")
            from .utils.usd_util import MdlRandomizer

            self.table_randomizer = MdlRandomizer(
                "/World/envs/env_0/metasim_table", case="table", split=self.scenario.split
            )
            self.ground_randomizer = MdlRandomizer("/World/ground", case="ground", split=self.scenario.split)
            self.wall_randomizer = MdlRandomizer(
                "/World/envs/env_0/metasim_wall", case="wall", split=self.scenario.split
            )

        for camera in self.cameras:
            ## Randomize camera pose
            if self.scenario.random.camera:
                from .utils.camera_random_util import randomize_camera_pose

                ## Get object in interest
                ## XXX: temporarily choose the first object as object in interest
                obj_pos, obj_quat = get_pose(env, self.objects[0].name)
                obj_pos = obj_pos[0]  # FIXME: only support one environment

                ## Get robot
                robot_pos, robot_quat = get_pose(env, self.robot.name)
                robot_quat = robot_quat[0]  # FIXME: only support one environment

                camera = randomize_camera_pose(camera, obj_pos.tolist(), robot_quat.tolist(), "front_select", self.task)

            if self.first_reset or self.scenario.random.camera:
                eyes = torch.tensor(camera.pos, dtype=torch.float32, device=env.device)[None, :]
                targets = torch.tensor(camera.look_at, dtype=torch.float32, device=env.device)[None, :]
                eyes = eyes + env.scene.env_origins
                targets = targets + env.scene.env_origins
                camera_inst = env.scene.sensors[camera.name]
                camera_inst.set_world_poses_from_view(eyes=eyes, targets=targets)

                ## XXX: update config to make sure alignment
                camera.pos = eyes[0].tolist()
                camera.look_at = targets[0].tolist()

        ## Randomize ground
        if self.scenario.random.ground:
            self.ground_randomizer.do_it()

        ## Randomize wall
        if self.scenario.random.wall:
            self.wall_randomizer.do_it()

        ## Randomize table
        if self.scenario.random.table and self.task.can_tabletop and self.scenario.try_add_table:
            from omni.isaac.core.prims import GeometryPrim

            from metasim.utils.math import quat_from_matrix

            self.table_randomizer.do_it()

            for env_id in range(env.num_envs):
                table_prim = GeometryPrim(f"/World/envs/env_{env_id}/metasim_table", name=f"table_prim_{env_id}")
                pos, _ = table_prim.get_world_pose()
                theta = torch.rand(1) * 2 * torch.pi
                rot_mat = torch.tensor([
                    [torch.cos(theta), -torch.sin(theta), 0.0],
                    [torch.sin(theta), torch.cos(theta), 0.0],
                    [0.0, 0.0, 1.0],
                ])
                quat = quat_from_matrix(rot_mat)
                table_prim.set_world_pose(position=pos, orientation=quat)

        ## Randomize objects reflection
        if self.scenario.random.reflection:
            from .utils.usd_util import ReflectionRandomizer

            for env_id in range(env.num_envs):
                for obj in self.objects:
                    if isinstance(obj, PrimitiveSphereMetaCfg) or isinstance(obj, PrimitiveCubeMetaCfg):
                        continue
                    ReflectionRandomizer(f"/World/envs/env_{env_id}/{obj.name}").do_it()

            if self.task.can_tabletop and self.scenario.try_add_table:
                ReflectionRandomizer("/World/envs/env_0/metasim_table").do_it()
            ReflectionRandomizer("/World/ground").do_it()

        ## Randomize lights
        if self.scenario.random.light:
            import numpy as np
            import omni.isaac.core.utils.prims as prim_utils
            from pxr import Gf

            nrow = 2
            ncol = 3

            light_radius = np.random.uniform(0.05, 0.1)
            light_length = np.random.uniform(0.5, 2.0)
            light_intensity = np.random.uniform(4e4, 1.2e5)
            light_color_temperature = np.random.uniform(4000, 8000)

            for row in range(nrow):
                for col in range(ncol):
                    light_prim = prim_utils.get_prim_at_path(f"/World/envs/env_0/lights/light_{row}_{col}")

                    light_spacing = (1.0, 2.5 * light_length)
                    light_prim.GetAttribute("xformOp:translate").Set((
                        light_spacing[1] * (col - (ncol - 1) / 2),
                        light_spacing[0] * (row - (nrow - 1) / 2),
                        wall_height - (table_height + table_thickness) - wall_thickness,
                    ))
                    light_prim.GetAttribute("xformOp:orient").Set(Gf.Quatd(1.0, 0.0, 0.0, 0.0))
                    light_prim.GetAttribute("inputs:radius").Set(light_radius)
                    light_prim.GetAttribute("inputs:length").Set(light_length)
                    light_prim.GetAttribute("inputs:intensity").Set(light_intensity)
                    light_prim.GetAttribute("inputs:enableColorTemperature").Set(True)
                    light_prim.GetAttribute("inputs:colorTemperature").Set(light_color_temperature)
        ## Special cases
        if self.task.__class__.__name__ == "StackGreenCubeMetaCfg":
            from random import randint, uniform

            import omni.isaac.core.utils.prims as prim_utils
            from pxr import Gf, Sdf, UsdShade

            from metasim.utils.color_util import hsv_to_rgb

            random_green = hsv_to_rgb(randint(90, 140), uniform(0.2, 1.0), uniform(0.2, 0.65))
            shader = UsdShade.Shader(prim_utils.get_prim_at_path("/World/envs/env_0/cube/geometry/material/Shader"))
            shader.CreateInput("diffuseColor", Sdf.ValueTypeNames.Color3f).Set(Gf.Vec3f(*random_green))

            random_orange = hsv_to_rgb(randint(15, 30), uniform(0.6, 1.0), uniform(0.75, 1.0))
            shader = UsdShade.Shader(prim_utils.get_prim_at_path("/World/envs/env_0/base/geometry/material/Shader"))
            shader.CreateInput("diffuseColor", Sdf.ValueTypeNames.Color3f).Set(Gf.Vec3f(*random_orange))

        if self.first_reset:
            self.first_reset = False

    def _setup_scene(self, env: "EmptyEnv") -> None:
        import omni.isaac.lab.sim as sim_utils
        from omni.isaac.core.prims import GeometryPrim
        from omni.isaac.core.utils.stage import add_reference_to_stage
        from omni.isaac.lab.sensors import TiledCamera, TiledCameraCfg
        from omni.isaac.lab.sim.spawners import spawn_light
        from omni.isaac.lab.utils.math import quat_from_euler_xyz

        from .utils.ground_util import create_ground, set_ground_material, set_ground_material_scale
        from .utils.usd_util import ReflectionRandomizer, ShaderFixer

        use_scene = False
        if self.scene is not None or self.scenario.random.scene:
            if self.scenario.random.scene:
                from .utils.usd_util import SceneRandomizer

                scene_randomizer = SceneRandomizer()
                scene_cfg_dict = scene_randomizer.do_it()
            else:
                scene_cfg_dict = {
                    "filepath": self.scene.filepath,
                    "position": self.scene.default_position,
                    "quat": self.scene.quat,
                    "scale": self.scene.scale,
                }

            add_reference_to_stage(
                scene_cfg_dict["filepath"],
                SCENE_PRIM_PATH,
            )
            house = GeometryPrim(SCENE_PRIM_PATH)
            quat = scene_cfg_dict["quat"]
            house.set_local_scale(scene_cfg_dict["scale"])
            house.set_world_pose(position=scene_cfg_dict["position"], orientation=quat)
            ShaderFixer(scene_cfg_dict["filepath"], SCENE_PRIM_PATH).fix_all()
            use_scene = True

        add_robot(env, self.robot)
        add_objects(env, self.objects + self.checker_debug_viewers[:1])  # TODO: now only support one checker viewer

        ## Fix shader texture map path
        for obj in self.objects:
            if isinstance(obj, RigidObjMetaCfg) or isinstance(obj, ArticulationObjMetaCfg):
                fixer = ShaderFixer(obj.filepath, f"/World/envs/env_0/{obj.name}")
                fixer.fix_all()

        ## Add ground
        ## TODO: test if this will conflict with the scene
        if not use_scene:
            create_ground()

        ## Add table
        if self.task.can_tabletop and self.scenario.try_add_table:
            import omni.isaac.core.utils.prims as prim_utils
            from omni.isaac.core.prims import GeometryPrim

            from .utils.custom_cuboid import FixedCuboid
            from .utils.ground_util import GROUND_PRIM_PATH

            ## Move ground down
            ground_prim = GeometryPrim(GROUND_PRIM_PATH, name="ground_prim")
            ground_prim.set_world_pose(position=(0.0, 0.0, -table_height), orientation=(1.0, 0.0, 0.0, 0.0))

            ## Add table
            prim_utils.create_prim("/World/envs/env_0/metasim_table")
            FixedCuboid(
                prim_path="/World/envs/env_0/metasim_table/surface",
                name="fixed_table",
                scale=torch.tensor([table_size, table_size, table_thickness]),
                position=torch.tensor([0.0, 0.0, -table_thickness / 2]),
            )
            for i, (x, y) in enumerate([
                (-table_size * 3 / 8, -table_size * 3 / 8),
                (table_size * 3 / 8, -table_size * 3 / 8),
                (-table_size * 3 / 8, table_size * 3 / 8),
                (table_size * 3 / 8, table_size * 3 / 8),
            ]):
                FixedCuboid(
                    prim_path=f"/World/envs/env_0/metasim_table/leg_{i}",
                    name=f"fixed_table_leg_{i}",
                    scale=torch.tensor([0.05, 0.05, table_height - table_thickness]),
                    position=torch.tensor([x, y, -(table_height + table_thickness) / 2]),
                )
        ## Add wall
        if self.scenario.random.wall:
            import omni.isaac.core.utils.prims as prim_utils

            from .utils.custom_cuboid import FixedCuboid

            prim_utils.create_prim("/World/envs/env_0/metasim_wall")

            for i, (x, y, sx, sy) in enumerate([
                (0, wall_dist, wall_dist * 2, wall_thickness),
                (0, -wall_dist, wall_dist * 2, wall_thickness),
                (wall_dist, 0, wall_thickness, wall_dist * 2),
                (-wall_dist, 0, wall_thickness, wall_dist * 2),
            ]):
                FixedCuboid(
                    f"/World/envs/env_0/metasim_wall/wall_{i}",
                    name=f"wall_{i}",
                    scale=torch.tensor([sx, sy, wall_height]),
                    position=torch.tensor([x, y, wall_height / 2 - (table_height + table_thickness)]),
                )

            ## Add roof
            if self.scenario.random.light:
                FixedCuboid(
                    prim_path="/World/envs/env_0/metasim_wall/roof",
                    name="roof",
                    scale=torch.tensor([wall_dist * 2, wall_dist * 2, wall_thickness]),
                    position=torch.tensor([
                        0.0,
                        0.0,
                        wall_height + wall_thickness / 2 - (table_height + table_thickness),
                    ]),
                )

        ## Set default ground material
        if not use_scene:
            set_ground_material("metasim/data/quick_start/materials/Ash.mdl")
            set_ground_material_scale(10)

        ## Clone, filter, and replicate
        env.scene.clone_environments(copy_from_source=False)
        env.scene.filter_collisions(global_prim_paths=[])

        ## Randomize material
        ## XXX: First call must be in _setup_scene, otherwise it will raise strange error
        if self.scenario.random.reflection:
            for env_id in range(env.num_envs):
                for obj in self.objects:
                    if isinstance(obj, PrimitiveSphereMetaCfg) or isinstance(obj, PrimitiveCubeMetaCfg):
                        continue
                    ReflectionRandomizer(f"/World/envs/env_{env_id}/{obj.name}").do_it()

        ## Add lights
        if self.scenario.random.light:
            ## Randomization mode: use cylinder light
            import omni.isaac.core.utils.prims as prim_utils

            nrow = 2
            ncol = 3

            for row in range(nrow):
                for col in range(ncol):
                    prim_utils.create_prim(
                        f"/World/envs/env_0/lights/light_{row}_{col}",
                        "CylinderLight",
                    )
        else:
            ## Normal mode: use distant light
            ## Light direction: 45 degree down between X+ and Z+
            light_quat = quat_from_euler_xyz(
                torch.tensor([0.0]),
                torch.tensor([0.0]),
                torch.tensor([0.0]),
            )[0].tolist()
            env.light = spawn_light(
                "/World/Light",
                sim_utils.DistantLightCfg(
                    intensity=500.0,
                ),
                orientation=light_quat,
            )

        ## Add camera
        for camera in self.cameras:
            env.scene.sensors[camera.name] = TiledCamera(
                TiledCameraCfg(
                    prim_path=f"/World/envs/env_.*/{camera.name}",
                    offset=TiledCameraCfg.OffsetCfg(pos=(0.0, 0.0, 0.0), rot=(1.0, 0.0, 0.0, 0.0), convention="world"),
                    data_types=camera.data_types,
                    spawn=sim_utils.PinholeCameraCfg(
                        focal_length=camera.focal_length,
                        focus_distance=camera.focus_distance,
                        horizontal_aperture=camera.horizontal_aperture,
                        clipping_range=camera.clipping_range,
                    ),
                    width=camera.width,
                    height=camera.height,
                )
            )

    def _pre_physics_step(self, env: "EmptyEnv", actions: torch.Tensor) -> None:
        ## TODO: Clip action or not?
        env.actions = actions

    def _apply_action(self, env: "EmptyEnv") -> None:
        env.robot.set_joint_position_target(env.actions)

    def _get_observations(self, env: "EmptyEnv") -> None:
        ## TODO: get proprioception observations
        for camera in self.cameras:
            camera_inst = env.scene.sensors[camera.name]

            ## Vision
            rgb_data = camera_inst.data.output.get("rgb", None)
            depth_data = camera_inst.data.output.get("depth", None)

            ## Camera
            cam_pos = torch.tensor(camera.pos, device=env.device)[None, :].repeat(env.num_envs, 1)
            cam_look_at = torch.tensor(camera.look_at, device=env.device)[None, :].repeat(env.num_envs, 1)
            cam_intr0 = camera_inst.data.intrinsic_matrices
            cam_extr, cam_intr = get_cam_params(
                cam_pos,
                cam_look_at,
                width=camera.width,
                height=camera.height,
                focal_length=camera.focal_length,
                horizontal_aperture=camera.horizontal_aperture,
            )
            assert torch.allclose(cam_intr0, cam_intr)

            ## Robot State
            joint_qpos_target = env.actions.cpu()
            joint_qpos = env.robot.data.joint_pos.cpu()
            robot_root_state = env.robot.data.root_state_w.cpu()
            robot_body_state = env.robot.data.body_state_w.cpu()
            robot_root_state[..., 0:3] -= env.scene.env_origins.cpu()
            robot_body_state[..., 0:3] -= env.scene.env_origins.cpu().unsqueeze(1)
            if self.robot.ee_prim_path is not None:
                ee_pos, ee_quat = get_pose(env, self.robot.name, self.robot.ee_prim_path)
                robot_ee_state = torch.cat([ee_pos, ee_quat, joint_qpos[:, -1:]], dim=-1)
            else:
                log.warning(f"No end-effector prim path for {self.robot.name}")
                robot_ee_state = None

            data_dict = {
                ## Vision
                "rgb": rgb_data,
                "depth": depth_data,
                ## Camera
                "cam_pos": cam_pos,
                "cam_look_at": cam_look_at,
                "cam_intr": cam_intr,
                "cam_extr": cam_extr,
                ## State
                "joint_qpos_target": joint_qpos_target,
                "joint_qpos": joint_qpos,  # align with old version
                "robot_ee_state": robot_ee_state,  # align with old version
                "robot_root_state": robot_root_state,
                "robot_body_state": robot_body_state,
            }

            for k, v in data_dict.items():
                if isinstance(v, torch.Tensor):
                    data_dict[k] = v.cpu()
            torch.cuda.empty_cache()

            return data_dict

    def _get_rewards(self, env: "EmptyEnv") -> None:
        pass

    def _get_dones(self, env: "EmptyEnv") -> tuple[torch.Tensor, torch.Tensor]:
        time_out = env.episode_length_buf >= env.max_episode_length - 1
        is_success = torch.zeros_like(time_out)
        return is_success, time_out



import gymnasium as gym
import numpy as np
import rootutils
import torch
from gymnasium import spaces
from loguru import logger as log
from rich.logging import RichHandler
from stable_baselines3 import PPO
from stable_baselines3.common.vec_env import DummyVecEnv

rootutils.setup_root(__file__, pythonpath=True)
log.configure(handlers=[{"sink": RichHandler(), "format": "{message}"}])

from metasim.cfg.scenario import ScenarioMetaCfg
from metasim.constants import SimType
from metasim.sim.base import BaseSimHandler
from metasim.utils.setup_util import get_robot, get_sim_env_class, get_task


class MetaSimGymEnv(gym.Env):
    """Vectorized environment for MetaSim that supports parallel RL training"""

    def __init__(self, scenario: ScenarioMetaCfg, num_envs: int = 1, sim_type: SimType = SimType.ISAACLAB):
        super().__init__()

        self.num_envs = num_envs
        env_class = get_sim_env_class(sim_type)
        env = env_class(scenario, num_envs)
        self.handler: BaseSimHandler = env.handler

        # Get joint limits for action space
        joint_limits = scenario.robot.joint_limits
        self.action_space = spaces.Box(
            low=np.array([lim[0] for lim in joint_limits.values()]),
            high=np.array([lim[1] for lim in joint_limits.values()]),
            dtype=np.float32,
        )

        # Observation space: joint positions + end effector position
        self.observation_space = spaces.Box(
            low=-np.inf,
            high=np.inf,
            shape=(len(joint_limits) + 3,),  # joints + XYZ
            dtype=np.float32,
        )

        self.max_episode_steps = 500
        self.current_step = 0

    def reset(self, seed=None):
        log.info("Resetting environments")
        """Reset all environments"""
        self.current_step = 0
        init_states = self._get_default_states()
        self.handler.set_states(init_states)
        self.handler.reset()
        return self._get_obs(), None

    def step(self, actions):
        """Step through all environments"""
        self.current_step += 1

        # Convert numpy actions to handler format
        ## TODO: action should support vectorized actions
        action_dicts = [{"dof_pos_target": dict(zip(self.handler.robot.joint_limits.keys(), actions))}]

        # Step the simulation
        self.handler.step(action_dicts)

        # Get new observations and calculate rewards
        obs = self._get_obs()
        rewards = self._calculate_rewards()
        timeout = np.array([self.current_step >= self.max_episode_steps] * self.num_envs)
        success = np.zeros_like(timeout)

        return obs, rewards, success, timeout, {}

    def _get_obs(self):
        """Get current observations for all environments"""
        states = self.handler.get_states()
        joint_pos = np.array([
            [state[self.handler.robot.name]["dof_pos"][j] for j in self.handler.robot.joint_limits.keys()]
            for state in states
        ])

        # Get end effector positions (assuming 'ee' is the end effector subpath)
        ee_pos = np.array([state["metasim_body_panda_hand"]["pos"] for state in states])

        return np.concatenate([joint_pos, ee_pos], axis=1)

    def _calculate_rewards(self):
        """Calculate rewards based on distance to origin"""
        states = self.handler.get_states()
        ee_pos = np.array([state["metasim_body_panda_hand"]["pos"] for state in states])
        distances = np.linalg.norm(ee_pos, axis=1)
        return -distances  # Negative distance as reward

    def _get_default_states(self):
        """Generate default reset states"""
        return [
            {
                self.handler.robot.name: {
                    "dof_pos": {j: 0.0 for j in self.handler.robot.joint_limits.keys()},
                    "pos": torch.tensor([0.0, 0.0, 0.0]),
                    "rot": torch.tensor([1.0, 0.0, 0.0, 0.0]),
                }
            }
            for _ in range(self.num_envs)
        ]

    def close(self):
        self.handler.close()


def train_ppo():
    """Training procedure for PPO"""
    # Environment setup
    scenario = ScenarioMetaCfg(
        task=get_task("PickCube"),
        robot=get_robot("franka_stable"),
    )

    env = MetaSimGymEnv(scenario, num_envs=1)
    env = DummyVecEnv([lambda: env])

    # PPO configuration
    model = PPO(
        "MlpPolicy",
        env,
        verbose=1,
        learning_rate=3e-4,
        n_steps=2048,
        batch_size=64,
        n_epochs=10,
        gamma=0.99,
        gae_lambda=0.95,
        clip_range=0.2,
        device="cuda" if torch.cuda.is_available() else "cpu",
    )

    # Start training
    model.learn(total_timesteps=1_000_000)
    model.save("ppo_reach")


if __name__ == "__main__":
    train_ppo()



from __future__ import annotations

import argparse
import logging
import os
import time

import numpy as np
import rootutils
import tqdm
from loguru import logger as log
from rich.logging import RichHandler

rootutils.setup_root(__file__, pythonpath=True)

logging.addLevelName(5, "TRACE")
log.configure(handlers=[{"sink": RichHandler(), "format": "{message}"}])


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--task", type=str, required=True)
    parser.add_argument("--robot", type=str, default="franka")
    parser.add_argument("--num_envs", type=int, default=1, help="Number of environments to simulate.")
    parser.add_argument("--num_trajs", type=int, default=10, help="Number of trajectories to play.")
    parser.add_argument(
        "--sim", type=str, default="isaaclab", choices=["isaaclab", "isaacgym", "pyrep", "pybullet", "sapien"]
    )
    parser.add_argument("--object_states", action="store_true", help="Replay object states.")
    args = parser.parse_args()
    return args


###########################################################
## Utils
###########################################################
def get_actions(all_actions, action_idx: int, num_envs: int):
    envs_actions = all_actions[:num_envs]
    actions = [
        env_actions[action_idx] if action_idx < len(env_actions) else env_actions[-1] for env_actions in envs_actions
    ]
    return actions


def get_states(all_states, action_idx: int, num_envs: int):
    envs_states = all_states[:num_envs]
    states = [env_states[action_idx] if action_idx < len(env_states) else env_states[-1] for env_states in envs_states]
    return states


def save_obs(obs, step: int):
    from torchvision.utils import make_grid, save_image

    if obs is None:
        return

    rgb_data = obs["rgb"]  # (N, H, W, C), uint8
    image = make_grid(rgb_data.permute(0, 3, 1, 2) / 255, nrow=int(rgb_data.shape[0] ** 0.5))  # (C, H, W)
    os.makedirs("tmp", exist_ok=True)
    save_image(image, f"tmp/rgb_all_{step:04d}.png")


###########################################################
## Main
###########################################################
def main():
    args = parse_args()
    num_envs: int = args.num_envs

    # specificly for isaacgym
    if args.sim == "isaacgym":
        from isaacgym import gymapi, gymtorch, gymutil  # noqa: F401

    ## Import put here to support isaacgym

    from metasim.cfg.cameras import PinholeCameraMetaCfg
    from metasim.cfg.scenario import ScenarioMetaCfg
    from metasim.constants import SimType
    from metasim.utils.demo_util import get_traj
    from metasim.utils.setup_util import get_robot, get_sim_env_class, get_task

    gym_env_class = get_sim_env_class(SimType(args.sim))
    task = get_task(args.task)
    robot = get_robot(args.robot)
    camera = PinholeCameraMetaCfg(
        name="camera",
        data_types=["rgb", "depth"],
        width=256,
        height=256,
        pos=(1.5, 0.0, 1.5),
        look_at=(0.0, 0.0, 0.0),
    )
    scenario = ScenarioMetaCfg(task=task, robot=robot, cameras=[camera])
    gym_env = gym_env_class(scenario, num_envs)

    ## Data
    tic = time.time()
    assert os.path.exists(task.traj_filepath), f"Trajectory file: {task.traj_filepath} does not exist."
    init_states, all_actions, all_states = get_traj(task, robot, gym_env)
    init_states = init_states[: args.num_trajs]
    all_actions = all_actions[: args.num_trajs]
    all_states = all_states[: args.num_trajs]
    toc = time.time()
    log.trace(f"Time to load data: {toc - tic:.2f}s")

    ########################################################
    ## Main
    ########################################################

    success_list = []
    episode_success_list = []

    with tqdm.tqdm(total=len(init_states)) as pbar:
        for init_state, action_list, state_list in zip(init_states, all_actions, all_states):
            ## Reset before first step
            if args.sim == "pyrep":
                log.warning("IsaacGym does not support getting init states, skipping...")
            else:
                gym_env.handler.set_states([init_state])

            ## Main loop
            step = 0
            episode_success = []

            for action, state in zip(action_list, state_list):
                if args.object_states:
                    if args.sim == "isaacgym":
                        raise NotImplementedError("IsaacGym does not support object states")
                    if all_states is None:
                        raise ValueError("All states are None, please check the trajectory file")
                    gym_env.handler.render()
                    gym_env.handler.set_states([state])

                    ## XXX: hackdemo_idx
                    success = gym_env.handler.task.checker.check(gym_env.handler)

                else:
                    obs, reward, success, time_out, extras = gym_env.step([action])

                    episode_success.append(success)
                    success_list.append(success)

            episode_success = np.stack(episode_success)
            episode_success = np.any(episode_success, axis=0)
            episode_success_list.append(episode_success)

            pbar.update(1)

    success_list = np.stack(success_list)
    episode_success_list = np.stack(episode_success_list)

    avg_success = np.mean(success_list, axis=0)
    avg_episode_success = np.mean(episode_success_list, axis=0)

    print("Avg success:", avg_success)
    print("Avg episode success", avg_episode_success)


if __name__ == "__main__":
    main()



from __future__ import annotations

from metasim.utils import configclass

from .base_robot_metacfg import BaseActuatorMetaCfg, BaseRobotMetaCfg


@configclass
class SawyerMetaCfg(BaseRobotMetaCfg):
    name: str = "sawyer"
    num_joints: int = 10
    filepath: str = "roboverse_data/robots/sawyer/usd/sawyer_v2.usd"
    enabled_gravity: bool = False
    enabled_self_collisions: bool = False
    actuators: dict[str, BaseActuatorMetaCfg] = {
        "right_j0": BaseActuatorMetaCfg(),
        "right_j1": BaseActuatorMetaCfg(),
        "right_j2": BaseActuatorMetaCfg(),
        "right_j3": BaseActuatorMetaCfg(),
        "right_j4": BaseActuatorMetaCfg(),
        "right_j5": BaseActuatorMetaCfg(),
        "right_j6": BaseActuatorMetaCfg(),
        "head_pan": BaseActuatorMetaCfg(),
        "right_gripper_l_finger_joint": BaseActuatorMetaCfg(is_ee=True),
        "right_gripper_r_finger_joint": BaseActuatorMetaCfg(is_ee=True),
    }
    joint_limits: dict[str, tuple[float, float]] = {
        "right_j0": (-3.05, 3.05),
        "right_j1": (-3.8094999, 2.2736),
        "right_j2": (-3.0425998, 3.0425998),
        "right_j3": (-3.0438999, 3.0438999),
        "right_j4": (-2.9760998, 2.9760998),
        "right_j5": (-2.9760998, 2.9760998),
        "right_j6": (-4.7123996, 4.7123996),
        "head_pan": (-5.095, 0.9064),
        "right_gripper_l_finger_joint": (0.0, 0.020833),
        "right_gripper_r_finger_joint": (-0.020833, 0),
    }
    ee_prim_path: str = "sawyer_right_hand"
    gripper_release_q = [0.020833, -0.020833]
    gripper_actuate_q = [0.0, 0.0]

    curobo_ref_cfg_name: str = "sawyer.yml"
    curobo_tcp_rel_pos: tuple[float, float, float] = [0.0, 0.0, 0.105]
    curobo_tcp_rel_rot: tuple[float, float, float] = [0.0, 0.0, 0.0]



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class PlayJengaMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/play_jenga/v2"
    objects = [
        RigidObjMetaCfg(
            name="Cuboid",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid0",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid1",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid2",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid3",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid4",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid5",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid6",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid7",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid8",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid9",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid10",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid11",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="Cuboid12",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="target_cuboid",
            filepath="roboverse_data/assets/rlbench/play_jenga/Cuboid/usd/Cuboid.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



import omni
import omni.isaac.core.utils.prims as prim_utils
import omni.isaac.lab.sim as sim_utils
from pxr import Sdf, UsdShade

from .material_util import apply_mdl_to_prim

GROUND_PRIM_PATH = "/World/ground"


def create_ground():
    # original file is http://omniverse-content-production.s3-us-west-2.amazonaws.com/Assets/Isaac/4.2/Isaac/Environments/Grid/Materials/Textures/Wireframe_blue.png
    cfg_ground = sim_utils.GroundPlaneCfg(
        physics_material=sim_utils.RigidBodyMaterialCfg(static_friction=1.0, dynamic_friction=1.0),
        color=(1.0, 1.0, 1.0),
    )
    cfg_ground.func(GROUND_PRIM_PATH, cfg_ground)


def set_ground_material_scale(scale: float):
    ground_prim = prim_utils.get_prim_at_path(GROUND_PRIM_PATH)
    material = UsdShade.MaterialBindingAPI(ground_prim).GetDirectBinding().GetMaterial()
    shader = UsdShade.Shader(omni.usd.get_shader_from_material(material, get_prim=True))
    # Correspond to Shader -> Inputs -> UV -> Texture Tiling (in Isaac Sim 4.2.0)
    shader.CreateInput("texture_scale", Sdf.ValueTypeNames.Float2).Set((scale, scale))


def set_ground_material(material_mdl_path: str):
    apply_mdl_to_prim(material_mdl_path, GROUND_PRIM_PATH)



from __future__ import annotations

from metasim.utils.configclass import configclass

from .base_scene_metacfg import SceneMetaCfg


@configclass
class TapwaterScene155MetaCfg(SceneMetaCfg):
    """Config class for tapwater scene"""

    name: str = "tapwater_155"
    filepath: str = "roboverse_data/scenes/arnold/155/layout.usd"
    positions: list[tuple[float, float, float]] = [
        (1.2855, -0.43902, -0.86244),
        (-1.40262, -1.28942, 0.0034),
        (-2.99159, 1.82148, 0.0034),
        (4.09718, -1.00533, 0.0034),
    ]  # XXX: only positions are randomized for now
    default_position: tuple[float, float, float] = (1.2855, -0.43902, -0.86244)
    quat: tuple[float, float, float, float] = (0.7071068, 0.7071068, 0.0, 0.0)
    scale: tuple[float, float, float] = (0.01, 0.01, 0.01)



# ruff: noqa: F401

from .balance_metacfg import BalanceMetaCfg
from .base_locomotion_env import HumanoidTaskMetaCfg
from .crawl_metacfg import CrawlMetaCfg
from .hurdle_metacfg import HurdleMetaCfg
from .maze_metacfg import MazeMetaCfg
from .pole_metacfg import PoleMetaCfg
from .push_metacfg import PushMetaCfg
from .run_metacfg import RunMetaCfg
from .sit_metacfg import SitMetaCfg
from .slide_metacfg import SlideMetaCfg
from .stair_metacfg import StairMetaCfg
from .stand_metacfg import StandMetaCfg
from .walk_metacfg import WalkMetaCfg
